//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Program        : Sorting with Interfaces
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
//
//
// Description:
//
//   Test Arrays.sort sorting Point with various Comparators. In violation
// of standard Java practices, I have put all classes implementing 
// Comparators in this file. Note that ReverseAComparator is in the
// course library. Finally, the ExtraInfo interface and the classes
// implementing it (passed print) allow printing extra information that
// illustrates whether the sorting works.
//
//   In PointXIncreasing, when comparing two Points with the same X
// coordinate, they compare according to their Y coordinates. That is,
// (10,3) is considered less than (10,10). The same holds for
// PointYIncreasing. 
//
// Known Bugs     : None
//
// Future Plans   : None
//
// Program History:
//   9/26/04: R. Pattis - Operational in Java for 15-200
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


import edu.cmu.cs.pattis.cs151xx.Prompt;
import edu.cmu.cs.pattis.cs151xx.ReverseAComparator;

import java.util.Arrays;
import java.util.Comparator;

import java.awt.Point;


public class Application {


//////////////////////////////////////////////////

public interface ExtraInfo {
  public String extra (Point p);
}

public static class NoInfo implements ExtraInfo {
  public String extra (Point p)
  {return "";}
}

public static class DistanceInfo implements ExtraInfo {
  public String extra (Point p)
  {return " with distance = "+Math.sqrt(p.x*p.x+p.y+p.y);}
}

public static class AngleInfo implements ExtraInfo {
  public String extra (Point p)
  {return " with angle = "+Math.toDegrees(Math.atan2(p.y,p.x));}
}

//////////////////////////////////////////////////

public static class PointXIncreasing implements Comparator {
  public int compare (Object o1, Object o2)
  {
    Point p1 = (Point)o1;
    Point p2 = (Point)o2;
    int xcompare = p1.x-p2.x;
    if (xcompare != 0)
      return xcompare;
    else
      return p1.y-p2.y;
  }
}

public static class PointYIncreasing implements Comparator {
  public int compare (Object o1, Object o2)
  {
    Point p1 = (Point)o1;
    Point p2 = (Point)o2;
    int xcompare = p1.y-p2.y;
    if (xcompare != 0)
      return xcompare;
    else
      return p1.x-p2.x;
  }
}

public static class PointDistanceIncreasing implements Comparator {
  public int compare (Object o1, Object o2)
  {
    Point  p1 = (Point)o1;
    Point  p2 = (Point)o2;
    double d1 = Math.sqrt( p1.x*p1.x + p1.y+p1.y );
    double d2 = Math.sqrt( p2.x*p2.x + p2.y+p2.y );
    if      (d1 < d2) return -1;
    else if (d1 > d2) return +1;
    else              return  0;
    //return (int)(d1-d2) loses information
  }
}

public static class PointAngleIncreasing implements Comparator {
  public int compare (Object o1, Object o2)
  {
    Point  p1 = (Point)o1;
    Point  p2 = (Point)o2;
    double a1 = Math.toDegrees(Math.atan2(p1.y,p1.x));
    double a2 = Math.toDegrees(Math.atan2(p2.y,p2.x));
    if      (a1 < a2) return -1;
    else if (a1 > a2) return +1;
    else              return  0;
    //return (int)(d1-d2) loses information
  }
}

//////////////////////////////////////////////////




  public static void print (String message, Point[] p, ExtraInfo ei)
  {
    System.out.println(message);
    for (int i=0; i<p.length; i++)
      System.out.println("  "+p[i] + ei.extra(p[i]));
  }
 

 
  public static void makeNewArray()
  {
    int used = Prompt.forInt("  Enter size of array to create");
    original = new Point[used];
    
    for (int i=0; i<used; i++)
      original[i] = new Point((int)(Math.random()*100),
                              (int)(Math.random()*100));
  }


 


  //Present menu and get user's choice
  
  private static char menuPrompt ()
  {
    System.out.println("\n\nOptions");
    System.out.println("  n - New (random) Array");
    System.out.println("  x - Sort by X Coordinate (increasing)");
    System.out.println("  X - Sort by X Coordinate (decreasing)");
    System.out.println("  y - Sort by Y Coordinate (increasing)");
    System.out.println("  Y - Sort by Y Coordinate (decreasing)");
    System.out.println("  d - Sort by distance-from-origin (increasing)");
    System.out.println("  D - Sort by distance-from-origin (decreasing)");
    System.out.println("  a - Sort by angle-from-orign (increasing)");
    System.out.println("  A - Sort by angle-from-orign (decreasing)");
    System.out.println("  q - quit");
    
    return Prompt.forChar("\nEnter Command","nxXyYdDaAq");
  }
  
  
  
  ////////////////
  //Driver Program
  ////////////////
  
	public static void main(String[] args)
	{
    makeNewArray();	  
    print("Array is initialized as follows",original,new NoInfo());
    
    for (;;) {
      char selection = menuPrompt();
      if (selection == 'n') {
        makeNewArray();
        print("Array is now initialized as follows",original,new NoInfo());
      }else if (selection == 'x') {
        comparator = new PointXIncreasing();
        extraInfo  = new NoInfo();
      }else if (selection == 'X') {
        comparator = new ReverseAComparator(new PointXIncreasing());
        extraInfo  = new NoInfo();
      }else if (selection == 'y') {
        comparator = new PointYIncreasing();
        extraInfo  = new NoInfo();
      }else if (selection == 'Y') {
        comparator = new ReverseAComparator(new PointYIncreasing());
        extraInfo  = new NoInfo();
      }else if (selection == 'd') {
        comparator = new PointDistanceIncreasing();
        extraInfo  = new DistanceInfo();
      }else if (selection == 'D') {
        comparator = new ReverseAComparator(new PointDistanceIncreasing());
        extraInfo  = new DistanceInfo();
      }else if (selection == 'a') {
        comparator = new PointAngleIncreasing();
        extraInfo  = new AngleInfo();
      }else if (selection == 'A') {
        comparator = new ReverseAComparator(new PointAngleIncreasing());
        extraInfo  = new AngleInfo();
      }else if (selection == 'q')
        break;
        
      if (selection != 'n') {
        Arrays.sort(original,comparator);
        print("Array is sorted as follows",original,extraInfo);
      }
    }
  }
  
  
  //Fields (all static)
  private static Comparator comparator;
  private static ExtraInfo  extraInfo;
  private static Point[]    original;
  
  
  //Note, I could use anonymous classes: e.g.,
  //extraInfo = new ExtraInto{public String extra(Point p){return "";}};
  //instead of extraInfo = new NoInfo after declaring the NoInfo class.
}
